######################################################################
#:ClearQuest Web Java XML parser interface object
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2005, 2015.  All Rights Reserved. 
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

=head1 NAME

CQCC::CQWebJavaParser - XML Parser that uses XML::Parser to parse CQWebJava
API commands. 

This class provides a small number of public accessors to work with 
an external class providing XML to be parsed: 

ParserInit - Returns 1 if parser is initialized, else 0

ParserSendMessage - Returns 1 if message was sent successfully, else 0

ParserProcessXML - Returns status and properties hash list from parsing
message.

=head1 CLASS DESCRIPTION

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package CQCC::CQWebJavaParser
######################################################################

package CQCC::CQWebJavaParser;
use Exporter;
@ISA = ('Exporter');
@EXPORT = qw(&ParserInit &ParserDumpState &ParserSendMessage &ParserProcessXML);

use strict;
use XML::Parser;
use IO::Handle;
use LWP::UserAgent;
use IO::Socket::SSL;

use CQCC::IOBase;
use CQCC::Parm;

######################################################################
#::Class Globals 
######################################################################

my $singleton;        # Parser instance
my $user_agent;       # LWP user agent instance

my @stack = ();       # Stack of elements being processed
my $parentElement = ""; # Current parent element
my @elementList = (); # List of major elements processed during parsing
my %props = ();       # Internal storage for named properties

my $htmlActive = 0;   # Flag to indicate we are processing HTML message
my $htmlMessage = ""; # HTML message without formatting 

my @unexpectedElements;  # List of unexpected elements in message

############# Char data

my $charData = "";    # Accumulated character data from one or more calls

############# Query state information

my $QueryDBID;        # Query id used for CreateNewQuery, ExecuteQuery
my @colNames;         # Names of columns
my $colHeaders;       # Number of columns
my $colPtr;           # Next column (field) being acquired for record

my @row;              # Query rows, once completed
my $rowTotal;         # Total number of rows
my $rowBatchCount;    # Number of rows to load in a batch
my $rowPtr;           # Next row being read in
my @rowColumns;       # Current rows being worked on

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(@args)>

This class function uses sets up a local environment for testing;
calls IOClassFactory() to get an instance of the desired subclass; 
and invokes the object's SelfTest method for testing.

=back

=cut

######################################################################

sub ClassTest {
    my $filename = "CQWebJavaParser.xml";

    $| = 1;
    # Open test file
    # Loop: 
    # * read input until separator foun

    if (ParserInit(0) && open (INFILE, $filename)) { 
       my @answerList = ();
        my $answer;
        while ($answer = <INFILE>) {
            chomp $answer;
            # * when separator found
            # ** parse the input
            # ** dump the current state to Stdout including query results
            if ($answer =~ /RUN_TEST/) {
                my $doc = join("\n",@answerList);
                print "Document:\n" . $doc . "\n---------\n";
                ParserProcessXML(0, $doc);
                ParserDumpState();
                print "\n---------------------\n";
                @answerList = ();
            } else {
                push @answerList, $answer;
            }
        }
        close INFILE;
    }
    return;
}

sub InitAgent {
  my ($use_sslv3) = @_;
  if ($use_sslv3) {
    return LWP::UserAgent->new(
      ssl_opts => {
                   SSL_cipher_list => "ALL:!LOW:!RC4",
                  }
    );
  } else {
    return LWP::UserAgent->new(
      ssl_opts => {
                   SSL_cipher_list => "ALL:!LOW:!RC4",
                   SSL_version => "!SSLv2:!SSLv3" 
                  }
   );
  }
}

######################################################################
#::function ParserInit

=over 6

=item C<ParserInit(use_sslv3)>

Initializes a parser instance and returns 1 if successful. Used both
to initialize the local parser instance and to confirm that the parser
class can be used by a calling object. 

If use_sslv3 is nonzero, allow SSLv3 in the user agent

=back

=cut

######################################################################

sub ParserInit {
    my ($use_sslv3) = @_;
    if (! defined $singleton) { 
        $singleton  = 
            XML::Parser->new(Handlers => {
                Start => \&start_element,
                End => \&end_element,
                Char => \&characters,
                ExternEnt => \&noext,
                Doctype => \&nodoc,
                Unparsed => \&noentity
                });
        $user_agent = InitAgent($use_sslv3);
    }
    return $singleton ? 1 : 0;
}
 
######################################################################
#
# Stop expansion of external entity references.
#
######################################################################

sub noext {
  return undef;
}

######################################################################
#
# Stop expansion of other entities.
#
######################################################################

sub noentity {
  die("External entities not supported.");
}

sub nodoc {
  die("Document types not supported.");
}

######################################################################
#::function ParserProcessXML

=over 6

=item C<(status, propsHashRef) = ParserProcessXML(use_sslv3, documentString)>

Parses the provided string and returns the parsed data in a hash
reference for lookup by name. Names are in the form "element.property"
where properties are attributes or other stored information for the 
element. 

If use_sslv3 is nonzero, then initialize the LWP user agent to allow SSLv3.

The status code is 1 for success, 0 for failure. 

=back

=cut

######################################################################

sub ParserProcessXML {
    my ($use_sslv3, $docString) = @_;
    ParserInit($use_sslv3);

    # Initialize global state values
    @stack = ();
    @elementList = ();
    %props = ();
    @colNames = ();
    @row = ();
    @rowColumns = ();
    $rowPtr = 0;
    $props{"Response.Status"} = "Failure";
    @unexpectedElements = ();

    IODebugPrint("DETAILED", "ParserProcessXML\n");

    #TODO: Would be nice to handle this as a DTD handler instead
    # Turn DTD statement <?xml version=x encoding=x ?> into
    #    <xml version=x encoding=x></xml> element 
    if ($docString) { 
        my @lines = split(/^/, $docString);
        my $line;
        $docString = "";
        my $stop = 0;
        foreach $line (@lines) {
            if ($stop == 0) {
                if ($line =~ /\<\?xml([\s|\S]*)\?\>/) {
                    my $attrs = $1;
                    if ($attrs =~ /version='(\S*)'/) {
                        $props{"xml.version"} = $1;
                    }
                    if ($attrs =~ /encoding\=\'(\S*)\'/) {
                        $props{"xml.encoding"} = $1;
                    }
                    $line =~ s/\<\?xml([\s|\S]*)\?\>//;
                    $stop = 1;
                }
            }
            $docString .= $line; 
        }
    }
    eval {
        $singleton->parse($docString);
    };
    if ($@) {
        #TODO: Translate error message
        IOMessage("FATAL", ["CQWJP_E_XML_ERROR", $@]);
    }

    if ($#unexpectedElements > -1) {
        # Report first unexpected element only
        IOMessage("WARN", ["CQWJP_E_UNEXPECTED_ELEMENT", 
                            $unexpectedElements[0]]);
    }

    #Store list of elements into Response.ElementList
    $props{"Parser.ElementList"} = join(",", @elementList);

    return (1,\%props);
}

######################################################################
#::function ParserDumpState

=over 6

=item C<ParserDumpState()>

Debugging routine that dumps the current parser results properties. 

=back

=cut

######################################################################

sub ParserDumpState {
    print "dumpState\n";
    for (sort keys %props) {
        print "    $_=[" . $props{$_} . "]\n";
    }
    print "\n";
    print "QueryResults: \n";
    my $i;
    for ($i=0; $i < $rowPtr; $i++) {
        my @localRow = @{$row[$i]};
        print join(",", @localRow) . "\n";
    }
    print "\n";
}

######################################################################
#::function ParserSendMessage

=over 6

=item C<ParserSendMessage(http, host, port, url, cookie, message)>

Sends the message to the identified host at the given url and port
using the provided session cookie. The http parameter specifies "http"
for non secure connections and "https" for secure connections.

=back

=cut

######################################################################

sub ParserSendMessage {
    my ($http, $host, $port, $url, $cookie, $content, $use_sslv3) = @_;
    my $fullURL = "$http://$host" . 
                  ($port eq "80" ? "" : ":$port") . 
                  "$url";
     
    IODebugPrint("DETAILED", "ParserSendMessage - connecting to $fullURL\n");
    my $request = HTTP::Request->new('POST' => $fullURL); 

    $request->header( "Content-Length" => length($content));
    $request->header( "Content-Type" => "application/x-www-form-urlencoded");
    if ($cookie) {
        $request->header( "Cookie" => $cookie);
    }
    $request->content( $content);

    if (! defined $user_agent) {
        $user_agent = InitAgent($use_sslv3);
    }
    my $response = $user_agent->request($request);
    return $response;
}

######################################################################
#::XML::Parser entry points

sub start_element{ 
    my ($expat,$name,%atts ) = @_;
    $charData = "";
    my $properties = \%atts;
    processElement($expat,$name,$properties,"start");
}

sub end_element{ 
    my ($expat,$name,%atts ) = @_;
    my $properties = \%atts;
    processElement($expat,$name,$properties,"end");
}

# Append all character data into single string
sub characters {
    my ($expat,$text) = @_;
    $charData .= $text;
}

# Dummy routines 

sub start_document { }          # Do nothing
sub end_document { }            # Do nothing
sub processing_instruction { }  # Do nothing
sub comment { }                 # Do nothing
sub start_cdata { }             # Do nothing
sub end_cdata { }               # Do nothing
sub entity_reference { }        # Do nothing

######################################################################
#::function processElement

=over 6

=item C<processElement()>

This is the main parser logic which handles start_element and end_element
calls to store element attribute and character data and other element
processing logic and initialization. 

=back

=cut

######################################################################

sub processElement {
    my ($expat,$elementName,$properties,$operation) = @_;

    #############################################
    # General initialization

    if ($operation eq "end") {
        pop @stack;
    }
    $parentElement = top_stack();
    if ($operation eq "start") {
        push @stack, $elementName;
    }

    #############################################
    # HTML processing

    if ($elementName eq "html") {
        if ($operation eq "start") {
            $htmlActive = 1;
            $htmlMessage = "";
       }
        if ($operation eq "end") {
            $htmlActive = 0;
            $props{"Response.Status"} = "Failure";
            $props{"Response.ErrorCode"} = "Unknown";
            $props{"Error.Summary"} = "HTML response";
            my @lines = split("\n", $htmlMessage);
            my $msg = "";
            my $emptyLast = 0;
            for (@lines) {
                chomp $_;
                $_ =~ s/  / /g;
                my $emptyNow = ((length($_) eq 0) || $_ eq " ") ? 1 : 0;
                if ($emptyNow && $emptyLast) {
                    next;
                }
                $emptyLast = $emptyNow;
                $msg .= $_ . "\n";
            }         
            $props{"Error.Detail"} = $msg;
            push(@elementList,"Response");
            push(@elementList,"Error");
        }
    } 

    elsif ($htmlActive eq 1) {
        if ($operation eq "end") {
            $htmlMessage .= $charData;
            if (! ($charData =~ /\n/)) {
                $htmlMessage .= "\n";
            }
            $charData = "";
        }
    }
   
    #############################################
    # Response element 

    elsif ($elementName eq "Response") {
        if ($operation eq "start") {
            # Attribs: xmlns:xsi, xsi:noNamespaceSchemaLocation,
            #     ReqID, Status, ErrorCode
            setPropsFromAttribs($properties,$elementName);
            push(@elementList,"Response");
        }
    }

    #############################################
    # Error element

    elsif ($elementName eq "Error") {
        if ($operation eq "start") {
            # Attribs: ID
            setPropsFromAttribs($properties,$elementName);
            push(@elementList,"Error");
        }
    }

    elsif ($parentElement eq "Error") { 
        if ($elementName eq "Operation") {
            if ($operation eq "start") {
                # Attribs: ID
               setPropsFromAttribs($properties,"Error.$elementName");
            }
        }

        elsif ($elementName eq "Summary" || $elementName eq "Detail") {
            if ($operation eq "end") {
               $props{"Error.$elementName"} = $charData;
            }
        }
    }

    #############################################
    # Session element and subelements

    elsif ($elementName eq "Session") {
        if ($operation eq "start") {
            # Attribs: ID, *MultiSiteActivated, localeId, localeStr
            setPropsFromAttribs($properties, $elementName);
            push(@elementList,"Session");
        }
    }

    elsif ($parentElement eq "Session") {
        if ($elementName eq "StageLabel" ||
            $elementName eq "SuiteProdVer" ||
            $elementName eq "LDAP" ||
            $elementName eq "LDAPAuthenticated" ||
            $elementName eq "UserLoginName") {
            if ($operation eq "end") {
                $props{"Session.$elementName"} = $charData;
            }
        }
    }

    #############################################
    # QueryDef element
 
    elsif ($elementName eq "QueryDef") {
        # Attributes: ID, dbid, edname, SiteHasMastership
        if ($operation eq "start") {
            setPropsFromAttribs($properties, $elementName);
            push(@elementList,$elementName);
        }
    }

    #############################################
    # Resultset element

    elsif ($elementName eq "Resultset") {
        # Attributes: dbid, MultiSiteActivated,UserHasQueryModifyPrivilege,
        #    SiteHasMastership
        if ($operation eq "start") {
            setPropsFromAttribs($properties, $elementName);
        }
        elsif ($operation eq "end") {
            #Store query results rows into Resultset.Rows
            $props{"$elementName.Rows"} = \@row;
            push(@elementList,$elementName);
        }
    }

    elsif ($parentElement eq "Resultset" && 
           ($elementName eq "dbid" ||
            $elementName eq "edname" ||
            $elementName eq "Mode" || 
            $elementName eq "WSItemName")) {
        if ($operation eq "end") {
            $props{"ResultSet." . $elementName} = $charData;
        }
    }

    #############################################
    # Resultset column headers 

    elsif ($elementName eq "ColumnHeaders") {
        if ($operation eq "start") {
            # Attribs: Count
            setPropsFromAttribs($properties, $elementName);
            $colHeaders = $props{"$elementName.Count"};
            $colPtr = 0;
            #IODebugPrint("DETAILED", "ColumnHeaders: $colHeaders");
        }
        elsif ($operation eq "end") {
            $props{"$elementName.ColNames"} = \@colNames;
        }
    }

    elsif ($parentElement eq "ColumnHeaders") {
        if ($elementName eq "ColumnHeader") {
            # Attribs: Type (ignored for now)
            if ($operation eq "end") {
                $colNames[$colPtr] = $charData;
                $colPtr++;
            }
        }
    }

    #############################################
    # Resultset data rows

    elsif ($elementName eq "Rows") {
        if ($operation eq "start"){ 
            # Attribs: TotalCount, Count, StartRow
            setPropsFromAttribs($properties, $elementName);
        }
    }

    elsif ($parentElement eq "Rows") {
        if ($elementName eq "Row") {
            if ($operation eq "start") {
                # Attribs: dbid
                setPropsFromAttribs($properties, $elementName);
                @rowColumns = ($props{"$elementName.dbid"});
            }

            elsif ($operation eq "end") {
                my @localRow = @rowColumns;
                $row[$rowPtr] = \@localRow;
                #IODebugPrint("DETAILED", 
                #   "row[$rowPtr]: " . join(",", @{$row[$rowPtr]}) . "\n");
                $rowPtr++;
            }
        }
    }

    elsif ($parentElement eq "Row") {
        if ($elementName eq "Column") {
            if ($operation eq "end") {
                # Attribs: Type
                push(@rowColumns, $charData);
            }
        }
    }

    #############################################
    # Entity

    elsif ($elementName eq "Entity") {
        if ($operation eq "start") {
            # Attribs: ID, dbid, dn, edname, enCurAction, 
            #    enCurActionType, SiteHasMastership
            setPropsFromAttribs($properties, $elementName);
        }
    }

    elsif ($props{"Response.Status"} ne "Failure") {
        push @unexpectedElements, $elementName;
    }
}

######################################################################
#::Support routines 
######################################################################

sub top_stack {
    return $stack[ $#stack ];
}

sub setPropsFromAttribs {
    my ($attribsRef, $prefix) = @_;
    my %attribs = %{$attribsRef};
    for (keys %attribs) {
        my $v = $attribs{$_};
        my $n = $_;
        $props{"$prefix.$n"} = $v;
    }
}

######################################################################
#::Tail
######################################################################

#ClassTest();  #Temp only
1;
